from torchdata.datapipes.iter import IterDataPipe
from torchdata.datapipes import functional_datapipe
import xarray as xr
import numpy as np
import pandas as pd
import logging

_log = logging.getLogger(__name__)

@functional_datapipe("select_pv_systems_within_region")
class SelectPVSystemsWithinRegionIterDataPipe(IterDataPipe):
    def __init__(self, source_dp: IterDataPipe, location_dp: IterDataPipe, roi_width_km: float, roi_height_km: float):
        self.source_dp = source_dp
        self.location_dp = location_dp
        self.roi_width_km = roi_width_km
        self.roi_height_km = roi_height_km

    def __iter__(self):
        pass



from torchdata.datapipes.iter import IterDataPipe

import numpy as np
import xarray as xr


class ReduceNumPVSystemsIterDataPipe(IterDataPipe):
    def __init__(self, source_dp: IterDataPipe, requested_num_pv_systems: int):
        super().__init__()
        self.source_dp = source_dp
        self.requested_num_pv_systems = requested_num_pv_systems
        self.rng = np.random.default_rng()

    def __iter__(self) -> xr.DataArray:
        yield None



from .pv import OpenPVFromDBIterDataPipe as OpenPVFromDB
from .pv import OpenPVFromNetCDFIterDataPipe as OpenPVFromNetCDF
from .satellite import OpenSatelliteDataPipe as OpenSatellite
from .nwp import OpenNWPIterDataPipe as OpenNWP
from .gsp import OpenGSPIterDataPipe as OpenGSP

try:
    import rioxarray  # Rioxarray is sometimes a pain to install, so only load this if its installed
    from .topographic import OpenTopographyIterDataPipe as OpenTopography
except ImportError:
    print("Rioxarray is not installed, so not importing OpenTopography")
    pass



from torchdata.datapipes.iter import IterDataPipe
from torchdata.datapipes import functional_datapipe
import numpy as np

@functional_datapipe("location_picker")
class LocationPickerIterDataPipe(IterDataPipe):
    def __init__(self, source_dp: IterDataPipe):
        super().__init__()
        self.source_dp = source_dp

    def __iter__(self):
        for xr_dataset in self.source_dp:
            # Assumes all datasets have osgb coordinates for selecting locations
            # Pick 1 random location from the input dataset
            location_idx = np.random.randint(0, len(xr_dataset["x_osgb"]))
            location = (xr_dataset["x_osgb"][location_idx], xr_dataset["y_osgb"][location_idx])
            yield location



from torchdata.datapipes.iter import IterDataPipe
from torchdata.datapipes import functional_datapipe

import fsspec
from pathy import Pathy
from pyaml_env import parse_config


@functional_datapipe("open_config")
class OpenConfigurationIterDataPipe(IterDataPipe):
    def __init__(self, configuration_filename: str):
        self.configuration_filename = configuration_filename

    def __iter__(self):
        with fsspec.open(self.configuration_filename, mode="r") as stream:
            configuration = parse_config(data=stream)

        # TODO Load into Pydantic Configuration class

        while True:
            yield configuration



from .reduce_num_pv_systems import ReduceNumPVSystemsIterDataPipe as ReduceNumPVSystems
from .select_pv_systems_within_region import SelectPVSystemsWithinRegionIterDataPipe as SelectPVSystemsWithinRegion



from torchdata.datapipes.iter import IterDataPipe
from torchdata.datapipes import functional_datapipe
import datetime
import geopandas as gpd
import xarray as xr
import pandas as pd
import numpy as np


@functional_datapipe("open_gsp")
class OpenGSPIterDataPipe(IterDataPipe):
    def __init__(
        self,
        gsp_pv_power_zarr_path: str,
        gsp_id_to_region_id_filename: str,
        sheffield_solar_region_path: str,
        threshold_mw: int = 0,
        sample_period_duration: datetime.timedelta = datetime.timedelta(minutes=30),
    ):
        self.gsp_pv_power_zarr_path = gsp_pv_power_zarr_path
        self.gsp_id_to_region_id_filename = gsp_id_to_region_id_filename
        self.sheffield_solar_region_path = sheffield_solar_region_path
        self.threshold_mw = threshold_mw
        self.sample_period_duration = sample_period_duration

    def __iter__(self):
        gsp_id_to_shape = _get_gsp_id_to_shape(
            self.gsp_id_to_region_id_filename, self.sheffield_solar_region_path
        )
        self._gsp_id_to_shape = gsp_id_to_shape  # Save, mostly for plotting to check all is fine!

        # Load GSP generation xr.Dataset:
        gsp_pv_power_mw_ds = xr.open_dataset(self.gsp_pv_power_zarr_path, engine="zarr")

        # Ensure the centroids have the same GSP ID index as the GSP PV power:
        gsp_id_to_shape = gsp_id_to_shape.loc[gsp_pv_power_mw_ds.gsp_id]

        data_array = _put_gsp_data_into_an_xr_dataarray(
            gsp_pv_power_mw=gsp_pv_power_mw_ds.generation_mw.data.astype(np.float32),
            time_utc=gsp_pv_power_mw_ds.datetime_gmt.data,
            gsp_id=gsp_pv_power_mw_ds.gsp_id.data,
            # TODO: Try using `gsp_id_to_shape.geometry.envelope.centroid`. See issue #76.
            x_osgb=gsp_id_to_shape.geometry.centroid.x.astype(np.float32),
            y_osgb=gsp_id_to_shape.geometry.centroid.y.astype(np.float32),
            capacity_mwp=gsp_pv_power_mw_ds.installedcapacity_mwp.data.astype(np.float32),
            t0_idx=self.t0_idx,
        )

        del gsp_id_to_shape, gsp_pv_power_mw_ds
        while True:
            yield data_array


def _get_gsp_id_to_shape(
    gsp_id_to_region_id_filename: str, sheffield_solar_region_path: str
) -> gpd.GeoDataFrame:
    # Load mapping from GSP ID to Sheffield Solar region ID:
    gsp_id_to_region_id = pd.read_csv(
        gsp_id_to_region_id_filename,
        usecols=["gsp_id", "region_id"],
        dtype={"gsp_id": np.int64, "region_id": np.int64},
    )

    # Load Sheffield Solar region shapes (which are already in OSGB36 CRS).
    ss_regions = gpd.read_file(sheffield_solar_region_path)

    # Merge, so we have a mapping from GSP ID to SS region shape:
    gsp_id_to_shape = (
        ss_regions.merge(gsp_id_to_region_id, left_on="RegionID", right_on="region_id")
        .set_index("gsp_id")[["geometry"]]
        .sort_index()
    )

    # Some GSPs are represented by multiple shapes. To find the correct centroid,
    # we need to find the spatial union of those regions, and then find the centroid
    # of those spatial unions. `dissolve(by="gsp_id")` groups by "gsp_id" and gets
    # the spatial union.
    return gsp_id_to_shape.dissolve(by="gsp_id")


def _put_gsp_data_into_an_xr_dataarray(
    gsp_pv_power_mw: np.ndarray,
    time_utc: np.ndarray,
    gsp_id: np.ndarray,
    x_osgb: np.ndarray,
    y_osgb: np.ndarray,
    capacity_mwp: np.ndarray,
    t0_idx: int,
) -> xr.DataArray:
    # Convert to xr.DataArray:
    data_array = xr.DataArray(
        gsp_pv_power_mw,
        coords=(("time_utc", time_utc), ("gsp_id", gsp_id)),
        name="gsp_pv_power_mw",
    )
    data_array = data_array.assign_coords(
        x_osgb=("gsp_id", x_osgb),
        y_osgb=("gsp_id", y_osgb),
        capacity_mwp=(("time_utc", "gsp_id"), capacity_mwp),
    )
    data_array.attrs["t0_idx"] = t0_idx
    return data_array



import datetime

from pathlib import Path
from typing import Optional, Union

from torchdata.datapipes.iter import IterDataPipe
from torchdata.datapipes import functional_datapipe
import logging

import fsspec
import numpy as np
import pandas as pd
import xarray as xr

_log = logging.getLogger(__name__)


@functional_datapipe("open_pv_netcdf")
class OpenPVFromNetCDFIterDataPipe(IterDataPipe):
    def __init__(
        self,
        pv_power_filename: str,
        pv_metadata_filename: str,
        n_pv_systems_per_example: int,
        sample_period_duration: datetime.timedelta = datetime.timedelta(minutes=5),
    ):
        super().__init__()
        self.pv_power_filename = pv_power_filename
        self.pv_metadata_filename = pv_metadata_filename
        self.n_pv_systems_per_example = n_pv_systems_per_example
        self.sample_period_duration = sample_period_duration

    def __iter__(self):
        data: xr.DataArray = load_everything_into_ram(
            self.pv_power_filename, self.pv_metadata_filename, self.sample_period_duration
        )
        while True:
            yield data


@functional_datapipe("open_pv_from_db")
class OpenPVFromDBIterDataPipe(IterDataPipe):
    def __init__(self):
        super().__init__()

    def __iter__(self):
        pass


def load_everything_into_ram(
    pv_power_filename, pv_metadata_filename, sample_period_duration
) -> xr.DataArray:
    """Open AND load PV data into RAM."""
    # Load pd.DataFrame of power and pd.Series of capacities:
    pv_power_watts, pv_capacity_wp, pv_system_row_number = _load_pv_power_watts_and_capacity_wp(
        pv_power_filename,
    )
    pv_metadata = _load_pv_metadata(pv_metadata_filename)
    # Ensure pv_metadata, pv_power_watts, and pv_capacity_wp all have the same set of
    # PV system IDs, in the same order:
    pv_metadata, pv_power_watts = _intersection_of_pv_system_ids(pv_metadata, pv_power_watts)
    pv_capacity_wp = pv_capacity_wp.loc[pv_power_watts.columns]
    pv_system_row_number = pv_system_row_number.loc[pv_power_watts.columns]

    data_in_ram = _put_pv_data_into_an_xr_dataarray(
        pv_power_watts=pv_power_watts,
        y_osgb=pv_metadata.y_osgb.astype(np.float32),
        x_osgb=pv_metadata.x_osgb.astype(np.float32),
        capacity_wp=pv_capacity_wp,
        pv_system_row_number=pv_system_row_number,
        sample_period_duration=sample_period_duration,
    )

    # Sanity checks:
    time_utc = pd.DatetimeIndex(data_in_ram.time_utc)
    assert time_utc.is_monotonic_increasing
    assert time_utc.is_unique

    return data_in_ram


def _load_pv_power_watts_and_capacity_wp(
    filename: Union[str, Path],
    start_date: Optional[datetime.datetime] = None,
    end_date: Optional[datetime.datetime] = None,
) -> tuple[pd.DataFrame, pd.Series, pd.Series]:
    """Return pv_power_watts, pv_capacity_wp, pv_system_row_number.

    The capacities and pv_system_row_number are computed across the *entire* dataset,
    and so is independent of the `start_date` and `end_date`. This ensures the PV system
    row number and capacities stay constant across training and validation.
    """

    _log.info(f"Loading solar PV power data from {filename} from {start_date=} to {end_date=}.")

    # Load data in a way that will work in the cloud and locally:
    with fsspec.open(filename, mode="rb") as file:
        pv_power_ds = xr.open_dataset(file, engine="h5netcdf")
        pv_capacity_wp = pv_power_ds.max().to_pandas().astype(np.float32)
        pv_power_watts = pv_power_ds.sel(datetime=slice(start_date, end_date)).to_dataframe()
        pv_power_watts = pv_power_watts.astype(np.float32)
        del pv_power_ds

    pv_capacity_wp.index = [np.int32(col) for col in pv_capacity_wp.index]

    # Create pv_system_row_number. We use the index of `pv_capacity_wp` because that includes
    # the PV system IDs for the entire dataset (independent of `start_date` and `end_date`).
    # We use `float32` for the ID because we use NaN to indicate a missing PV system,
    # or that this whole example doesn't include PV.
    all_pv_system_ids = pv_capacity_wp.index
    pv_system_row_number = np.arange(start=0, stop=len(all_pv_system_ids), dtype=np.float32)
    pv_system_row_number = pd.Series(pv_system_row_number, index=all_pv_system_ids)

    _log.info(
        "After loading:"
        f" {len(pv_power_watts)} PV power datetimes."
        f" {len(pv_power_watts.columns)} PV power PV system IDs."
    )

    # Sanity checks:
    assert not pv_power_watts.columns.duplicated().any()
    assert not pv_power_watts.index.duplicated().any()
    assert np.isfinite(pv_capacity_wp).all()
    assert (pv_capacity_wp > 0).all()
    assert np.isfinite(pv_system_row_number).all()
    assert np.array_equal(pv_power_watts.columns, pv_capacity_wp.index)
    return pv_power_watts, pv_capacity_wp, pv_system_row_number


"""Filtering to be added in a different IterDataPipe

    pv_power_watts = pv_power_watts.clip(lower=0, upper=5e7)
    # Convert the pv_system_id column names from strings to ints:
    pv_power_watts.columns = [np.int32(col) for col in pv_power_watts.columns]

    if "passiv" not in filename:
        _log.warning("Converting timezone. ARE YOU SURE THAT'S WHAT YOU WANT TO DO?")
        pv_power_watts = (
            pv_power_watts.tz_localize("Europe/London").tz_convert("UTC").tz_convert(None)
        )

    pv_power_watts = _drop_pv_systems_which_produce_overnight(pv_power_watts)

    # Resample to 5-minutely and interpolate up to 15 minutes ahead.
    # TODO: Issue #74: Give users the option to NOT resample (because Perceiver IO
    # doesn't need all the data to be perfectly aligned).
    pv_power_watts = pv_power_watts.resample("5T").interpolate(method="time", limit=3)
    pv_power_watts.dropna(axis="index", how="all", inplace=True)
    pv_power_watts.dropna(axis="columns", how="all", inplace=True)

    # Drop any PV systems whose PV capacity is too low:
    PV_CAPACITY_THRESHOLD_W = 100
    pv_systems_to_drop = pv_capacity_wp.index[pv_capacity_wp <= PV_CAPACITY_THRESHOLD_W]
    pv_systems_to_drop = pv_systems_to_drop.intersection(pv_power_watts.columns)
    _log.info(
        f"Dropping {len(pv_systems_to_drop)} PV systems because their max power is less than"
        f" {PV_CAPACITY_THRESHOLD_W}"
    )
    pv_power_watts.drop(columns=pv_systems_to_drop, inplace=True)

    # Ensure that capacity and pv_system_row_num use the same PV system IDs as the power DF:
    pv_system_ids = pv_power_watts.columns
    pv_capacity_wp = pv_capacity_wp.loc[pv_system_ids]
    pv_system_row_number = pv_system_row_number.loc[pv_system_ids]

    _log.info(
        "After filtering & resampling to 5 minutes:"
        f" pv_power = {pv_power_watts.values.nbytes / 1e6:,.1f} MBytes."
        f" {len(pv_power_watts)} PV power datetimes."
        f" {len(pv_power_watts.columns)} PV power PV system IDs."
    )


"""
# Adapted from nowcasting_dataset.data_sources.pv.pv_data_source
def _load_pv_metadata(filename: str) -> pd.DataFrame:
    """Return pd.DataFrame of PV metadata.

    Shape of the returned pd.DataFrame for Passiv PV data:
        Index: ss_id (Sheffield Solar ID)
        Columns: llsoacd, orientation, tilt, kwp, operational_at,
            latitude, longitude, system_id, x_osgb, y_osgb
    """
    _log.info(f"Loading PV metadata from {filename}")
    pv_metadata = pd.read_csv(filename, index_col="ss_id").drop(columns="Unnamed: 0")
    _log.info(f"Found {len(pv_metadata)} PV systems in {filename}")

    # drop any systems with no lon or lat:
    pv_metadata.dropna(subset=["longitude", "latitude"], how="any", inplace=True)

    _log.debug(f"Found {len(pv_metadata)} PV systems with locations")
    return pv_metadata


def _intersection_of_pv_system_ids(
    pv_metadata: pd.DataFrame, pv_power: pd.DataFrame
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Only pick PV systems for which we have metadata."""
    pv_system_ids = pv_metadata.index.intersection(pv_power.columns)
    pv_system_ids = np.sort(pv_system_ids)
    pv_power = pv_power[pv_system_ids]
    pv_metadata = pv_metadata.loc[pv_system_ids]
    return pv_metadata, pv_power


def _put_pv_data_into_an_xr_dataarray(
    pv_power_watts: pd.DataFrame,
    y_osgb: pd.Series,
    x_osgb: pd.Series,
    capacity_wp: pd.Series,
    pv_system_row_number: pd.Series,
    sample_period_duration: datetime.timedelta,
) -> xr.DataArray:
    """Convert to an xarray DataArray.

    Args:
        pv_power_watts: pd.DataFrame where the columns are PV systems (and the column names are
            ints), and the index is UTC datetime.
        x_osgb: The x location. Index = PV system ID ints.
        y_osgb: The y location. Index = PV system ID ints.
        capacity_wp: The max power output of each PV system in Watts. Index = PV system ID ints.
        pv_system_row_number: The integer position of the PV system in the metadata.
            Used to create the PV system ID embedding.
    """
    # Sanity check!
    pv_system_ids = pv_power_watts.columns
    for series in (x_osgb, y_osgb, capacity_wp, pv_system_row_number):
        assert np.array_equal(series.index, pv_system_ids, equal_nan=True)

    data_array = xr.DataArray(
        data=pv_power_watts.values,
        coords=(("time_utc", pv_power_watts.index), ("pv_system_id", pv_power_watts.columns)),
        name="pv_power_watts",
    ).astype(np.float32)

    data_array = data_array.assign_coords(
        x_osgb=("pv_system_id", x_osgb),
        y_osgb=("pv_system_id", y_osgb),
        capacity_wp=("pv_system_id", capacity_wp),
        pv_system_row_number=("pv_system_id", pv_system_row_number),
    )
    # Sample period duration is required so PVDownsample transform knows by how much
    # to change the pv_t0_idx:
    data_array.attrs["sample_period_duration"] = sample_period_duration
    return data_array



from torchdata.datapipes.iter import IterDataPipe, Zipper
from torchdata.datapipes import functional_datapipe
from typing import Union, Iterable


@functional_datapipe("select_overlapping_time_slice")
class SelectOverlappingTimeSlice(IterDataPipe):
    def __init__(self, source_dps: Iterable[IterDataPipe]):
        super().__init__()
        self.source_dps = source_dps

    def __iter__(self):
        for set_of_xrs in Zipper(*self.source_dps):
            pass



from torchdata.datapipes.iter import IterDataPipe
from torchdata.datapipes import functional_datapipe
import xarray as xr
import numpy as np
from ocf_datapipes.utils import NumpyBatch
from ocf_datapipes.consts import BatchKey

@functional_datapipe("add_topographic_data")
class AddTopographicDataIterDataPipe(IterDataPipe):
    def __init__(self, source_dp: IterDataPipe, topo_dp: IterDataPipe):
        self.source_dp = source_dp
        self.topo_dp = topo_dp

    def __iter__(self):
        for topo in self.topo_dp:
            for np_batch in self.source_dp:
                if BatchKey.hrvsatellite_x_geostationary in np_batch:
                    # Recreate an xr.DataArray of the satellite data. This is required so we can
                    # use xr.combine_by_coords to align the topo data with the satellite data.
                    hrvsatellite_data_array = xr.DataArray(
                        # We're not actually interested in the image. But xarray won't make an
                        # empty DataArray without data in the right shape.
                        # There's nothing special about the x_osgb data. It's just convenient because
                        # it's the right shape!
                        np_batch[BatchKey.hrvsatellite_x_osgb],
                        dims=("example", "y", "x"),
                        coords={
                            "y_geostationary": (
                                ("example", "y"),
                                np_batch[BatchKey.hrvsatellite_y_geostationary],
                            ),
                            "x_geostationary": (
                                ("example", "x"),
                                np_batch[BatchKey.hrvsatellite_x_geostationary],
                            ),
                        },
                    )
                    hrvsatellite_surface_height = _get_surface_height_for_satellite(
                        surface_height=topo, satellite=hrvsatellite_data_array
                    )
                    np_batch[BatchKey.hrvsatellite_surface_height] = hrvsatellite_surface_height
                yield np_batch


def _get_surface_height_for_satellite(
    surface_height: xr.DataArray, satellite: xr.DataArray
) -> np.ndarray:
    num_examples = satellite.shape[0]
    surface_height = surface_height.rename("surface_height")
    surface_height_for_batch = np.full_like(satellite.values, fill_value=np.NaN)
    for example_idx in range(num_examples):
        satellite_example = satellite.isel(example=example_idx)
        msg = "Satellite imagery must start in the top-left!"
        assert satellite_example.y_geostationary[0] > satellite_example.y_geostationary[-1], msg
        assert satellite_example.x_geostationary[0] < satellite_example.x_geostationary[-1], msg

        satellite_example = satellite_example.rename(
            {"y_geostationary": "y", "x_geostationary": "x"}
        ).rename("sat")
        surface_height_for_example = surface_height.sel(
            y=slice(
                satellite_example.y[0],
                satellite_example.y[-1],
            ),
            x=slice(
                satellite_example.x[0],
                satellite_example.x[-1],
            ),
        )
        # Align by coordinates. This will result in lots of NaNs in the surface height data:
        aligned = xr.combine_by_coords(
            (surface_height_for_example, satellite_example), join="outer"
        )
        # Fill in the NaNs:
        surface_height_for_example = (
            aligned["surface_height"].ffill("x").ffill("y").bfill("x").bfill("y")
        )

        # Now select exactly the same coordinates from the surface height data as the satellite data
        aligned = xr.combine_by_coords((surface_height_for_example, satellite_example), join="left")

        surface_height_for_batch[example_idx] = aligned["surface_height"].values

    # If we slightly ran off the edge of the topo data then we'll get NaNs.
    # TODO: Enlarge topo data so we never get NaNs!
    surface_height_for_batch = np.nan_to_num(surface_height_for_batch, nan=0)

    return surface_height_for_batch



from torchdata.datapipes.iter import IterDataPipe
from torchdata.datapipes import functional_datapipe
import xarray as xr
from typing import Union
from pathlib import Path
import logging
import numpy as np
import dask
import pandas as pd

_log = logging.getLogger(__name__)


@functional_datapipe("open_nwp")
class OpenNWPIterDataPipe(IterDataPipe):
    def __init__(self, zarr_path: Union[Path, str]):
        self.zarr_path = zarr_path

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        _log.debug("Opening NWP data: %s", self.zarr_path)
        nwp = xr.open_dataset(
            self.zarr_path,
            engine="zarr",
            consolidated=True,
            mode="r",
            chunks="auto",
        )

        ukv: xr.DataArray = nwp["UKV"]
        del nwp

        ukv = ukv.transpose("init_time", "step", "variable", "y", "x")
        ukv = ukv.rename(
            {"init_time": "init_time_utc", "variable": "channel", "y": "y_osgb", "x": "x_osgb"}
        )

        # y_osgb and x_osgb are int64 on disk.
        for coord_name in ("y_osgb", "x_osgb"):
            ukv[coord_name] = ukv[coord_name].astype(np.float32)

        # Sanity checks.
        assert ukv.y_osgb[0] > ukv.y_osgb[1], "UKV must run from top-to-bottom."
        time = pd.DatetimeIndex(ukv.init_time_utc)
        assert time.is_unique
        assert time.is_monotonic_increasing
        while True:
            yield ukv
